"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const typescript_operations_1 = require("@graphql-codegen/typescript-operations");
const visitor_plugin_common_1 = require("@graphql-codegen/visitor-plugin-common");
const plugin_helpers_1 = require("@graphql-codegen/plugin-helpers");
const graphql_1 = require("graphql");
const autoBind = require("auto-bind");
const typescript_1 = require("@graphql-codegen/typescript");
/**
 * A function returning the named type of a variable or field
 *
 * @param variableDef A definition node containing a NamedType in its structure
 */
const getNamedTypeName = (variableDef) => { var _a, _b; return variableDef.kind === 'NamedType' ? (_b = (_a = variableDef.name) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : '' : getNamedTypeName(variableDef.type); };
// noinspection JSUnusedGlobalSymbols
/**
 * A visitor class to extract all TypeDefinition nodes with a given name.
 *
 * This works by passing a string array with the required types and on each InputObject-, Object- and
 *  InterfaceTypeDefinition their name is looked up in that array. If it's required it remains in the ast, otherwise
 *  it is removed by returning null.
 *
 * Enum and Scalar definitions are excluded as they are already defined in a "global" scope.
 * Union definitions are excluded as they guaranteed not to be used and only result in broken types across all operations.
 */
class TypeCollectorVisitor {
    constructor(typesToInclude) {
        this.typesToInclude = typesToInclude;
        autoBind(this);
    }
    DirectiveDefinition() {
        return null;
    }
    DirectiveTypeDefinition() {
        return null;
    }
    EnumTypeDefinition() {
        return null;
    }
    ScalarTypeDefinition() {
        return null;
    }
    UnionTypeDefinition() {
        return null;
    }
    /**
     * Filtering function that returns the input node or null depending on whether its name is in the typesToInclude or
     * not
     *
     * @param node The node to check
     */
    _ifTypeIncluded(node) {
        const name = node.name.value;
        return this.typesToInclude.includes(name) ? node : null;
    }
    InputObjectTypeDefinition(node) {
        return this._ifTypeIncluded(node);
    }
    InterfaceTypeDefinition(node) {
        return this._ifTypeIncluded(node);
    }
    ObjectTypeDefinition(node) {
        return this._ifTypeIncluded(node);
    }
}
/**
 * A custom extension of the real TsVisitor to generate types as defined in the concept:
 *
 * Scalar types are preresolved to avoid Scalar['String'] constructs
 * Nullable types are generated as T | null instead of Maybe<T>
 * Prepend the namespacedImportName (usually Types) to enum types from the "global" file in generated types.
 */
class CustomTsVisitor extends typescript_1.TsVisitor {
    constructor(schema, pluginConfig) {
        super(schema, pluginConfig);
        this.pluginConfig = pluginConfig;
        this.nullableSuffix = ' | null';
    }
    /**
     * Called by the TsVisitor to determine the string representation of a named scalar
     *
     * Original implementation returned `Scalar['${name}']`; this adaption resolves it from the internal map of scalars
     *
     * @param name Name of the scalar to get the type string for.
     * @param type Whether the scalar is an input or output scalar.
     */
    _getScalar(name, type) {
        return this.scalars[name][type] || super._getScalar(name, type);
    }
    /**
     * Called by the TsVisitor to clear the optional declaration from nodes that are NotNullTypes
     *
     * The original implementation removed Maybe<>; this is an adaption to the requirement for | null
     *
     * @param str String containing the typedefinition the optional should be removed from.
     */
    clearOptional(str) {
        if (str.endsWith(this.nullableSuffix)) {
            return str.substring(0, str.length - this.nullableSuffix.length);
        }
        return str;
    }
    /**
     * Called by the TsVisitor to transform a NamedType to its string representation
     *
     * The original implementation wrapped the result of a super call in Maybe<>; this adaption is a copy from that super
     * call's code adding the nullable suffix
     *
     * @param node NamedTypeNode the type string should be generated for.
     * @param isVisitingInputType Whether the Visitor is currently visiting an input type
     */
    NamedType(node, isVisitingInputType) {
        return this._getTypeForNode(node, isVisitingInputType) + this.nullableSuffix;
    }
    /**
     * Called by the TsVisitor to transform a ListType to its string representation
     *
     * The original implementation wrapped the result of a super call in Maybe<>; this adaption is a copy from that super
     * call's code adding the nullable suffix
     *
     * @param node ListTypeNode the type string should be generated for.
     */
    ListType(node) {
        const asString = node.type.toString();
        const listType = this.wrapWithListType(asString);
        return listType + this.nullableSuffix;
    }
    /**
     * Called by the TsVisitor to determine the flat typename for a node.
     * Extended to prepend Types. to enum types; rest is handled by the default implementation
     *
     * @param node The node the type should be determined for
     * @param isVisitingInputType Whether the Visitor is currently visiting an input type
     */
    _getTypeForNode(node, isVisitingInputType) {
        var _a;
        // The fixed signature of this is inaccurate. The values passed often (always?) don't a name.value field; just name
        const schemaType = this._schema.getType((_a = node.name.value) !== null && _a !== void 0 ? _a : node.name);
        if (schemaType && (0, graphql_1.isEnumType)(schemaType)) {
            return (this.pluginConfig.namespacedImportName + '.' + this.convertName(node, { useTypesPrefix: this.config.enumPrefix }));
        }
        return super._getTypeForNode(node, isVisitingInputType);
    }
}
/**
 * Class converting the operation variables of a mutation or query to an object
 *
 * This extension of the original implementation serves the following purposes:
 *
 * Remove Types. prefix of classes that are generated in local scope (like input types)
 * Add Types. prefix to enums
 * Pre resolve Scalars
 */
class CustomOperationVariablesToObject extends typescript_1.TypeScriptOperationVariablesToObject {
    _getBaseTypeNode(typeNode) {
        if (typeNode.kind === graphql_1.Kind.LIST_TYPE || typeNode.kind === graphql_1.Kind.NON_NULL_TYPE) {
            return this._getBaseTypeNode(typeNode.type);
        }
        return typeNode;
    }
    /**
     * Called by the transformer to transform an operation variable definition to a TS object field definition
     *
     * This is a copy from the original implementation with slight adjustments where needed to:
     *
     * Remove Types. prefix of classes that are generated in local scope (like input types)
     * Add Types. prefix to enums
     * Pre resolve Scalars
     *
     * @param variable An operation Variable Definition to transform to a string
     */
    transformVariable(variable) {
        var _a, _b;
        let typeValue;
        const prefix = this._namespacedImportName ? `${this._namespacedImportName}.` : '';
        const baseType = this._getBaseTypeNode(variable.type);
        const typeName = baseType.name.value;
        if (this._scalars[typeName]) {
            const scalar = this._scalars[typeName];
            typeValue = (_a = scalar.input) !== null && _a !== void 0 ? _a : scalar.output;
        }
        else if ((_b = this._enumValues[typeName]) === null || _b === void 0 ? void 0 : _b.sourceFile) {
            typeValue = this._enumValues[typeName].typeIdentifier || this._enumValues[typeName].sourceIdentifier || '';
        }
        else if (this._enumNames.includes(typeName)) {
            typeValue =
                prefix +
                    this._convertName(baseType, {
                        useTypesPrefix: this._enumPrefix,
                    });
        }
        else {
            typeValue = this._convertName(baseType);
        }
        const fieldName = this.getName(variable);
        const fieldType = this.wrapAstTypeWithModifiers(typeValue, variable.type);
        const hasDefaultValue = variable.defaultValue != null && typeof variable.defaultValue !== 'undefined';
        const isNonNullType = variable.type.kind === graphql_1.Kind.NON_NULL_TYPE;
        const formattedFieldString = this.formatFieldString(fieldName, isNonNullType, hasDefaultValue);
        const formattedTypeString = this.formatTypeString(fieldType, isNonNullType, hasDefaultValue);
        return `${formattedFieldString}: ${formattedTypeString}`;
    }
}
/**
 * Custom types processor to ensure that exported fields get to keep their name
 */
class CustomPreResolveTypesProcessor extends visitor_plugin_common_1.PreResolveTypesProcessor {
    constructor(processorConfig) {
        super(processorConfig);
        this.exportAliases = new Map();
    }
    registerExportAlias(parentSchemaType, fieldName, exportedTypeName) {
        if (!this.exportAliases.has(parentSchemaType)) {
            this.exportAliases.set(parentSchemaType, new Map());
        }
        this.exportAliases.get(parentSchemaType).set(fieldName, exportedTypeName);
    }
    transformPrimitiveFields(schemaType, fields) {
        var _a;
        const exportAliasMap = this.exportAliases.get(schemaType);
        const regularFields = fields.filter((x) => !(exportAliasMap === null || exportAliasMap === void 0 ? void 0 : exportAliasMap.has(x.fieldName)));
        const exportedFields = fields.filter((x) => exportAliasMap === null || exportAliasMap === void 0 ? void 0 : exportAliasMap.has(x.fieldName));
        const transformedPrimitiveFields = (_a = super.transformPrimitiveFields(schemaType, regularFields)) !== null && _a !== void 0 ? _a : [];
        const exportedNameAndType = exportedFields.map((x) => ({ name: x.fieldName, type: exportAliasMap.get(x.fieldName) }));
        return [...exportedNameAndType, ...transformedPrimitiveFields];
    }
}
/**
 * Transformer class for graphql selection sets.
 *
 * Most functions here are copied from the original implementation and slightly adapted because they don't seem to be
 *  designed for overwriting them.
 */
class CustomSelectionSetToObject extends visitor_plugin_common_1.SelectionSetToObject {
    constructor(processor, scalars, schema, convertName, getFragmentSuffix, loadedFragments, config, parentSchemaType, selectionSet) {
        super(processor, scalars, schema, convertName, getFragmentSuffix, loadedFragments, config, parentSchemaType, selectionSet);
        autoBind(this);
    }
    /**
     * Convenience function to create aliases for subtypes of interfaces
     *
     * @param type The named type the alias should be generated for
     * @param mainAlias The alias of the exported interface type field's export directive
     */
    static _getInterfaceAlias(type, mainAlias) {
        return `${mainAlias}_${type.name}`;
    }
    /**
     * Returns the export alias of a field node or throws if either the export directive or the exportName is missing.
     *
     * @param node FieldNode to get the export alias for
     */
    static _getExportedAlias(node) {
        var _a, _b;
        const exportDirective = (_a = node.directives) === null || _a === void 0 ? void 0 : _a.find((directive) => directive.name.value === 'export');
        if (!exportDirective) {
            throw new Error(`Couldn't find export directive when trying to find the exported alias. FieldNode name is ${node.name}`);
        }
        const nameArg = (_b = exportDirective.arguments) === null || _b === void 0 ? void 0 : _b.find((arg) => arg.name.value === 'exportName');
        if (!nameArg) {
            throw new Error(`Couldn't find exportName on export directive when trying to find the exported alias. FieldNode name is ${node.name}`);
        }
        return nameArg.value.value;
    }
    /**
     * Creates a new instance of the SelectionSetToObject for another selection set
     * @param parentSchemaType ParentSchemaType of the selection set
     * @param selectionSet selection set to be used as base selection set for the new object
     */
    createNext(parentSchemaType, selectionSet) {
        return new CustomSelectionSetToObject(this._processor, this._scalars, this._schema, this._convertName.bind(this), this._getFragmentSuffix.bind(this), this._loadedFragments, this._config, parentSchemaType, selectionSet);
    }
    /**
     * Converts an array of selections to a SelectionSetNode
     * @param selections Array of SelectionNodes to be set as selections of the SelectionSetNode
     */
    static _selectionsToSelectionSet(selections) {
        return {
            kind: graphql_1.Kind.SELECTION_SET,
            selections,
        };
    }
    buildSelectionSet(parentSchemaType, selectionNodes, options) {
        const isExportMarkedType = (selectionNode) => selectionNode.marker;
        const exported = selectionNodes.filter(isExportMarkedType);
        const forwarded = selectionNodes.filter((node) => !isExportMarkedType(node));
        for (const { fieldName, exportedTypeName } of exported) {
            this._processor.registerExportAlias(parentSchemaType, fieldName, exportedTypeName);
        }
        const cheekyTrick = exported.map(({ fieldName, exportedTypeName }) => ({ name: { value: fieldName, kind: graphql_1.Kind.NAME }, type: exportedTypeName, kind: graphql_1.Kind.FIELD }));
        return super.buildSelectionSet(parentSchemaType, [...forwarded, ...cheekyTrick], options);
    }
    /**
     * Flattens the selection set by trimming off irrelevant bits and converting fragments to a form that's easier to work
     *  with.
     *
     * This implementation manipulates the returned map of the super call by going through the fields and patching those
     *  that should be exported
     *
     * @param selections A list of SelectionNodes of a selection set
     */
    flattenSelectionSet(selections) {
        if (!this._parentSchemaType) {
            return new Map();
        }
        const isFieldNode = (node) => {
            return node.kind === "Field";
        };
        const map = super.flattenSelectionSet(selections);
        for (const [currentTypeName, fields] of map.entries()) {
            const fieldsWithExports = fields.map((field) => {
                if (!isFieldNode(field) ||
                    !field.directives ||
                    field.directives.every((directive) => directive.name.value !== 'export')) {
                    return field;
                }
                const exportUsage = this.buildExportUsage(field);
                const usage = exportUsage.get(currentTypeName);
                return usage !== null && usage !== void 0 ? usage : field;
            });
            map.set(currentTypeName, fieldsWithExports);
        }
        return map;
    }
    /**
     * Function to flatten the selection set on a given type that is not the base type
     *
     * @param selections A list of SelectionNodes of a selection set
     * @param type The type the selection is performed on
     */
    _flattenFromType(selections, type) {
        if (!this._parentSchemaType || (0, graphql_1.isEqualType)(type, this._parentSchemaType)) {
            return this.flattenSelectionSet(selections);
        }
        const subSelectionSet = this.createNext(type, CustomSelectionSetToObject._selectionsToSelectionSet(selections));
        return subSelectionSet.flattenSelectionSet(selections);
    }
    /**
     * Function converting a series of selectionNodes to a large selection set string in the form of a type definition.
     *
     * This implementation appends the exported field definitions to the result of a super call.
     *
     * @param parentSchemaType Parent Schema Type the selection is performed on
     * @param selectionNodes Selection nodes. A string indicates a FragmentSpread being used, an object with
     *  {marker: true} indicates an exported node, all other objects are SelectionNodes
     */
    buildSelectionSetString(parentSchemaType, selectionNodes) {
        var _a;
        const isExportMarkedType = (selectionNode) => selectionNode.marker;
        const exported = selectionNodes.filter(isExportMarkedType);
        const forwarded = selectionNodes.filter((node) => !isExportMarkedType(node));
        const superSelectionSet = super.buildSelectionSet(parentSchemaType, forwarded, {});
        const superBuildString = (_a = super.selectionSetStringFromFields(superSelectionSet.fields.filter((f) => !!f))) !== null && _a !== void 0 ? _a : '';
        if (exported.length === 0) {
            return superBuildString;
        }
        const isAppendable = superBuildString && superBuildString.endsWith('}');
        const exportedTransformed = (isAppendable ? ';' : '{') + exported.map((exp) => `${exp.fieldName}: ${exp.exportedTypeName}`).join(',') + '}';
        if (isAppendable) {
            return superBuildString.substring(0, superBuildString.length - 1) + exportedTransformed;
        }
        return this._processor.buildSelectionSetFromStrings([superBuildString, exportedTransformed]);
    }
    /**
     * Function building usages of the exported type.
     *
     * The result of this function is an object containing mappings from typenames of types defined in the schema to
     *  objects containing information on the exported field:
     *
     *  marker: true - used to indicate this as exported field in buildSelectionSetString
     *  fieldName: string - indicates the exported field's name
     *  exportedTypeName: string - value of the exportedName and name of the type that is generated from the selection
     *
     * @param exp An exported FieldNode
     */
    buildExportUsage(exp) {
        var _a, _b;
        const map = new Map();
        const usage = CustomSelectionSetToObject._getExportedAlias(exp);
        const schemaType = this._getExportedSchemaType(usage, true);
        const possibleTypesForExport = (0, visitor_plugin_common_1.getPossibleTypes)(this._schema, schemaType);
        for (let possibleType of possibleTypesForExport) {
            const selectedSchemaType = schemaType.getFields()[exp.name.value].type;
            const wrapped = this._processor.config.wrapTypeWithModifiers(usage, selectedSchemaType);
            if (map.has(possibleType.name)) {
                throw new Error(`Already set an export marked type name for ${possibleType.name}`);
            }
            map.set(possibleType.name, {
                marker: true,
                fieldName: (_b = (_a = exp.alias) === null || _a === void 0 ? void 0 : _a.value) !== null && _b !== void 0 ? _b : exp.name.value,
                exportedTypeName: wrapped,
            });
        }
        return map;
    }
    /**
     * Extension of the original function to generate the exported types with it.
     *
     * @param fragmentName fragmentName to be forwarded to the original function
     * @param fragmentSuffix fragmentSuffix to be forwarded to the original function
     * @param declarationBlockConfig declarationBlockConfig to be forwarded to the original function
     */
    transformFragmentSelectionSetToTypes(fragmentName, fragmentSuffix, declarationBlockConfig) {
        return (super.transformFragmentSelectionSetToTypes(fragmentName, fragmentSuffix, declarationBlockConfig) +
            '\n' +
            this.getExportedTypes());
    }
    /**
     * Returns a string containing concatenated declaration blocks of all exported types in this selection set.
     */
    getExportedTypes() {
        var _a, _b, _c;
        if (!this._selectionSet || !this._parentSchemaType) {
            return '';
        }
        const exported = this._getExportedSelectionSets(this._selectionSet, this._parentSchemaType);
        const blocks = {};
        for (const { node, type, alias } of exported) {
            const fieldNode = node;
            const selections = [...((_b = (_a = fieldNode.selectionSet) === null || _a === void 0 ? void 0 : _a.selections) !== null && _b !== void 0 ? _b : [])];
            const flattened = this._flattenFromType(selections, type);
            if (!blocks[alias]) {
                blocks[alias] = [];
            }
            if ((0, graphql_1.isInterfaceType)(type)) {
                blocks[alias].push(this._getInterfaceAliases(type, alias).join(' | '));
            }
            else {
                blocks[alias].push(this.buildSelectionSetString(type, (_c = flattened.get(type.name)) !== null && _c !== void 0 ? _c : []));
            }
        }
        return Object.keys(blocks)
            .map((typeName) => new visitor_plugin_common_1.DeclarationBlock({})
            .export()
            .asKind('type')
            .withName(typeName)
            .withContent(blocks[typeName].join(' & ')).string)
            .join('\n');
    }
    /**
     * Returns the schema type the exported field is selected on (parent = true) or the type the exported type is derived
     *  from (parent = false).
     *
     * @param alias Name of the field that is exported.
     * @param parent true if the parentSchemaType should be returned, false otherwise (default)
     */
    _getExportedSchemaType(alias, parent = false) {
        if (!this._selectionSet || !this._parentSchemaType) {
            throw new Error(`Could not find type for alias ${alias} - selectionSet or parentSchemaType not defined`);
        }
        const exportedSelections = this._getExportedSelectionSets(this._selectionSet, this._parentSchemaType);
        const correctMapping = exportedSelections.find((mapping) => mapping.alias === alias);
        if (!correctMapping) {
            throw new Error(`Could not find type for alias ${alias}`);
        }
        return parent ? correctMapping.parentType : correctMapping.type;
    }
    /**
     * Recursively returns a list of SelectionNodes that have the export directive on them. The result includes the type
     *  of the exported field, the parentType and the alias.
     *
     * @param set The SelectionSetNode to filter exported nodes from
     * @param parentType Type of the field the selection is made on
     */
    _getExportedSelectionSets(set, parentType) {
        return set
            ? set.selections
                .map((selection) => this._getExportedSelectionSetNodes(selection, parentType))
                .reduce((a, b) => [...a, ...b], [])
            : [];
    }
    /**
     * Returns a list of SelectionNodes that have the export directive on them including the passed node itself if it has
     *  the export directive on it. The result includes the type of the exported field, the parentType and the alias.
     *
     * @param node Node to get exported SelectionNodes from.
     * @param parentType Type of the field the exported node is a child of
     */
    _getExportedSelectionSetNodes(node, parentType) {
        var _a;
        if (node.kind === 'InlineFragment') {
            const typeName = (_a = node.typeCondition) === null || _a === void 0 ? void 0 : _a.name.value;
            if (!typeName) {
                throw new Error('Could not find parent typename for resolving inline spread exports');
            }
            const type = this._schema.getType(typeName);
            if (!type) {
                throw new Error('Could not find parent type for resolving inline spread exports');
            }
            return this._getExportedSelectionSets(node.selectionSet, type);
        }
        if (node.kind !== 'Field' || (!(0, graphql_1.isObjectType)(parentType) && !(0, graphql_1.isInterfaceType)(parentType))) {
            return [];
        }
        const thisNodeField = parentType.getFields()[node.name.value];
        if (!thisNodeField) {
            return [];
        }
        const thisNodeType = (0, plugin_helpers_1.getBaseType)(thisNodeField.type);
        const subExportedFields = (0, graphql_1.isObjectType)(thisNodeType) || (0, graphql_1.isInterfaceType)(thisNodeType)
            ? this._getExportedSelectionSets(node.selectionSet, thisNodeType)
            : [];
        const hasExport = node.directives && node.directives.some((directive) => directive.name.value === 'export');
        if (hasExport && (0, graphql_1.isObjectType)(thisNodeType)) {
            return [
                { node, type: thisNodeType, parentType, alias: CustomSelectionSetToObject._getExportedAlias(node) },
                ...subExportedFields,
            ];
        }
        if (hasExport && (0, graphql_1.isInterfaceType)(thisNodeType)) {
            const subTypes = (0, visitor_plugin_common_1.getPossibleTypes)(this._schema, thisNodeType);
            const mainAlias = CustomSelectionSetToObject._getExportedAlias(node);
            const interfaceExports = subTypes.map((type) => ({ node, type, parentType, alias: `${mainAlias}_${type.name}` }));
            return [{ node, type: thisNodeType, parentType, alias: mainAlias }, ...interfaceExports, ...subExportedFields];
        }
        if (hasExport) {
            throw new Error(`Type ${thisNodeType.name} is a primitive and may not be exported! Field name is ${node.name.value}`);
        }
        return subExportedFields;
    }
    /**
     * Generates a string array containing the names of types of exported interface types.
     *
     * An interface can have multiple types that implement it; thus when a field of an interface type is exported, we have
     *  to generate a type for each type that implements the interface.
     *
     * @param type The interface type the aliases should be generated for.
     * @param mainAlias The exportName of the interface type field's export annotation.
     */
    _getInterfaceAliases(type, mainAlias) {
        return (0, visitor_plugin_common_1.getPossibleTypes)(this._schema, type).map((possibleType) => CustomSelectionSetToObject._getInterfaceAlias(possibleType, mainAlias));
    }
}
/**
 * A custom extension of the visitor that generates the typescript operations code to generate code as defined in the
 *  concept. The key element here is setting a custom variable's transformer.
 */
class CustomTypeScriptOperationsVisitor extends typescript_operations_1.TypeScriptDocumentsVisitor {
    constructor(schema, config, allFragments) {
        super(schema, config, allFragments);
        const wrapOptional = (type) => {
            return `${type} | null`;
        };
        const wrapArray = (type) => {
            const listModifier = this.config.immutableTypes ? 'ReadonlyArray' : 'Array';
            return `${listModifier}<${type}>`;
        };
        const formatNamedField = (name, type) => {
            const optional = !this.config.avoidOptionals.field && !!type && !(0, graphql_1.isNonNullType)(type);
            return (this.config.immutableTypes ? `readonly ${name}` : name) + (optional ? '?' : '');
        };
        const processorConfig = {
            namespacedImportName: this.config.namespacedImportName,
            convertName: this.convertName.bind(this),
            enumPrefix: this.config.enumPrefix,
            enumSuffix: this.config.enumSuffix,
            scalars: this.scalars,
            formatNamedField,
            wrapTypeWithModifiers(baseType, type) {
                return (0, visitor_plugin_common_1.wrapTypeWithModifiers)(baseType, type, { wrapOptional, wrapArray });
            },
        };
        const processor = new CustomPreResolveTypesProcessor(processorConfig);
        this.setSelectionSetHandler(new CustomSelectionSetToObject(processor, this.scalars, this.schema, this.convertName.bind(this), this.getFragmentSuffix.bind(this), allFragments, this.config));
        const enumsNames = Object.keys(schema.getTypeMap()).filter((typeName) => (0, graphql_1.isEnumType)(schema.getType(typeName)));
        this.setVariablesTransformer(new CustomOperationVariablesToObject(this.scalars, this.convertName.bind(this), !!this.config.avoidOptionals.object, this.config.immutableTypes, this.config.namespacedImportName, enumsNames, this.config.enumPrefix, this.config.enumSuffix, this.config.enumValues));
    }
    /**
     * Copy of a non-exported util function. Returns the schema root type for a given operation
     *
     * @param operation Operation name. One of 'query', 'mutation', 'subscription'.
     * @param schema The schema the root type is going to be looked up from.
     */
    static _getRootType(operation, schema) {
        switch (operation) {
            case 'query':
                return schema.getQueryType();
            case 'mutation':
                return schema.getMutationType();
            case 'subscription':
                return schema.getSubscriptionType();
        }
    }
    /**
     * Extension of the original OperationDefinition function to lookup and generate exported types
     *
     * @param node The OperationDefinitionNode that is the current target of AST-Visiting
     */
    OperationDefinition(node) {
        const operationRootType = CustomTypeScriptOperationsVisitor._getRootType(node.operation, this._schema);
        if (!operationRootType) {
            throw new Error(`Unable to find root schema type for operation type "${node.operation}"!`);
        }
        const selectionSet = this._selectionSetToObject.createNext(operationRootType, node.selectionSet);
        const exported = selectionSet.getExportedTypes();
        return super.OperationDefinition(node) + '\n' + exported;
    }
}
/**
 * Returns an array of typenames referenced by requiredTypes
 *
 * This method recursively includes types referenced by other types.
 *
 * @param schema GraphQL schema that was passed to the plugin
 * @param requiredTypes A string list of required types
 * @param previouslyReferenced An array containing already referenced types
 */
const getSubTypeNames = (schema, requiredTypes, previouslyReferenced = []) => {
    if (requiredTypes.length === 0) {
        return [];
    }
    const fieldedGraphQLTypeFilter = (type) => (0, graphql_1.isObjectType)(type) || (0, graphql_1.isInterfaceType)(type) || (0, graphql_1.isInputObjectType)(type);
    const schemaTypes = requiredTypes.map((type) => schema.getType(type)).filter(fieldedGraphQLTypeFilter);
    const subFields = schemaTypes.map((type) => Object.values(type.getFields())).reduce((a, b) => [...a, ...b], []);
    const subTypes = subFields.map((field) => (0, plugin_helpers_1.getBaseType)(field.type)).filter(fieldedGraphQLTypeFilter);
    const subTypeNames = subTypes.map((type) => type.name);
    const newPrevious = [...requiredTypes, ...previouslyReferenced];
    const newRequired = subTypeNames.filter((name) => !newPrevious.includes(name));
    const foundTypes = [...subTypeNames, ...getSubTypeNames(schema, newRequired, newPrevious)];
    return [...new Set(foundTypes)];
};
/**
 * Returns a list of all type names referenced by operations in a DocumentNode
 *
 * @param node DocumentNode containing the operations from which types should be extracted
 */
const getRequiredTypeNames = (node) => {
    var _a;
    const operations = node.definitions.filter((def) => def.kind === 'OperationDefinition');
    if (!operations) {
        return [];
    }
    const allVariableDefinitions = (_a = operations
        .map((operation) => operation.variableDefinitions)
        .filter((defs) => !!defs)
        .reduce((a, b) => [...a, ...b], [])) === null || _a === void 0 ? void 0 : _a.filter((def) => def.kind === 'VariableDefinition');
    if (!allVariableDefinitions) {
        return [];
    }
    const typeNames = allVariableDefinitions.map((def) => getNamedTypeName(def));
    return [...new Set(typeNames)];
};
/**
 * Function generating a list of lines containing the type definitions of the types required for this document
 *
 * @param documents List of documents to generate the types for
 * @param ast The full schema AST
 * @param schema The schema passed to the plugin to pass to the TsVisitor
 * @param pluginConfig The plugin config passed to the plugin to pass to the TsVisitor
 */
function getTypeContent(documents, ast, schema, pluginConfig) {
    // Note: This is not perfect because requiredTypeNames can contain scalars and enums, but they are thrown out
    //  in the process of searching type definitions either way.
    const requiredTypeNames = documents
        .map((document) => document.document)
        .filter((document) => !!document)
        .map((document) => getRequiredTypeNames(document))
        .reduce((a, b) => [...a, ...b]);
    const allTypeNames = [...requiredTypeNames, ...getSubTypeNames(schema, requiredTypeNames)];
    const astReducer = new TypeCollectorVisitor(allTypeNames);
    // @ts-expect-error Mismatch between graphql and graphql-codegen
    const reducedAst = (0, plugin_helpers_1.oldVisit)(ast, { leave: astReducer });
    const tsVisitor = new CustomTsVisitor(schema, pluginConfig);
    // @ts-expect-error Mismatch between graphql and graphql-codegen
    const generatedDefinitions = (0, plugin_helpers_1.oldVisit)(reducedAst, { leave: tsVisitor });
    return generatedDefinitions.definitions.join('\n');
}
const plugin = (schema, rawDocuments, config) => {
    var _a;
    const documents = config.flattenGeneratedTypes ? (0, visitor_plugin_common_1.optimizeOperations)(schema, rawDocuments) : rawDocuments;
    const allAst = (0, graphql_1.concatAST)(documents.map((v) => v.document).filter((d) => !!d));
    const printed = (0, graphql_1.printSchema)(schema);
    const node = (0, graphql_1.parse)(printed);
    const generatedTs = getTypeContent(documents, node, schema, config);
    const allFragments = [
        ...allAst.definitions.filter((d) => d.kind === graphql_1.Kind.FRAGMENT_DEFINITION).map((fragmentDef) => ({
            node: fragmentDef,
            name: fragmentDef.name.value,
            onType: fragmentDef.typeCondition.name.value,
            isExternal: false,
        })),
        ...(config.externalFragments || []),
    ];
    const visitor = new CustomTypeScriptOperationsVisitor(schema, config, allFragments);
    // @ts-expect-error Mismatch between graphql and graphql-codegen
    const result = (0, plugin_helpers_1.oldVisit)(allAst, { leave: visitor });
    const imports = (_a = config.fragmentImports) === null || _a === void 0 ? void 0 : _a.map((i) => (0, visitor_plugin_common_1.generateFragmentImportStatement)(i, 'type')).join('\n');
    return { prepend: [imports], content: [generatedTs, result.definitions.join('\n')].join('\n') };
};
const pluginInformation = { plugin };
module.exports = pluginInformation;
